import os
import subprocess

# Function to run a program with a given input file
def run_program(program, input_file):
    with open(input_file, "r") as fin:
        result = subprocess.run(program, stdin=fin, stdout=subprocess.PIPE, text=True)
    return result.stdout.strip()

# Function to compile the given source file
def compile_program(source_file, output_file):
    compile_command = ["g++", source_file, "-o", output_file, "--std=c++20"]
    result = subprocess.run(compile_command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
    if result.returncode != 0:
        print(f"Compilation failed for {source_file}. Errors:\n{result.stderr}")
        return False
    return True

# Main function that manages the compilation and testing
def main():
    # Directories and program names
    tests_dir = "tests"
    author_prog = "./author/author"
    author_fast_prog = "./author/author_fast"
    slow_10p_prog = "./author/slow_10p"
    slow_30p_prog = "./author/slow_30p"
    greedy_prog = "./author/greedy"

    # Compile the necessary programs
    print("Compiling the programs...")
    
    # List of C++ files to compile
    programs = [
        (f"{prog}.cpp", prog)
        for prog in [
            author_prog,
            author_fast_prog,
            slow_10p_prog,
            slow_30p_prog,
            greedy_prog,
        ]
    ]
    
    for source_file, output_file in programs:
        if not compile_program(source_file, output_file):
            return  # If compilation fails, stop further execution

    # Now run the tests
    for test_file in sorted(os.listdir(tests_dir)):
        if not test_file.endswith(".in"):
            continue
        
        # Paths to input and expected output files
        input_path = os.path.join(tests_dir, test_file)
        output_path = input_path.replace(".in", ".out")
        
        # Read the expected output
        with open(output_path, "r") as fout:
            expected_output = fout.read().strip()
        
        # Run the programs and get the outputs
        author_output = run_program(author_prog, input_path)
        slow_10p_output = run_program(slow_10p_prog, input_path)
        slow_30p_output = run_program(slow_30p_prog, input_path)
        greedy_output = run_program(greedy_prog, input_path)

        # Compare the outputs
        if author_output != expected_output or slow_10p_output != expected_output or slow_30p_output != expected_output:
            print(f"Test {test_file} failed!")
            print(f"Expected: {expected_output}")
            print(f"Author:   {author_output}")
            print(f"Slow 10p: {slow_10p_output}")
            print(f"Slow 30p: {slow_30p_output}")
        elif greedy_output == expected_output:
            print(f"Test {test_file} passed for correct solutions, but greedy unexpectedly passed.")
        else:
            print(f"Test {test_file} passed successfully.")

if __name__ == "__main__":
    main()
