import random
import subprocess
import json
import os
import math

# Function to generate the input test case
def generate_test_case(n, max_height, pattern=None):
    if pattern == "constant":
        return [max_height] * n
    elif pattern == "pyramid":
        return generate_pyramids(n)
    elif pattern == "stacked-pyramid":
        return generate_stacked_pyramids(n)
    else:
        return [random.randint(1, max_height) for _ in range(n)]

def generate_pyramids(n):
    pyramids = []
    remaining = n

    choices = [max(x, 1) for x in [1, int(math.sqrt(n)), n - 3, n // 2, n // 3]]
    p = random.choice(choices)

    while remaining > 0:
        max_size = min(remaining, p)
        height = random.randint(1, max_size)
        if height % 2 == 0:
            height -= 1  # Ensure height is odd
        pyramids.append(create_pyramid(height))
        remaining -= height

    return [block for pyramid in pyramids for block in pyramid]

def create_pyramid(height):
    if height == 1:
        return [1]
    half = (height // 2)
    return list(range(1, half + 2)) + list(range(half, 0, -1))

def generate_stacked_pyramids(n):
    arr = [1] * n
    remaining = n
    while remaining > 0:
        pyramid_height = random.randint(1, remaining)
        if pyramid_height % 2 == 0:
            pyramid_height -= 1
        start_idx = random.randint(0, n - pyramid_height)
        pyramid = create_pyramid(pyramid_height)
        for i in range(pyramid_height):
            arr[start_idx + i] = max(arr[start_idx + i], pyramid[i])
        remaining -= pyramid_height
    return arr

def load_file_content(file_path):
    with open(file_path, "r") as f:
        return f.read().strip()

def modify_test_case(original_test, max_height):
    # Convert test case to list of integers
    original_list = list(map(int, original_test.split()))
    # Chop off a random number of elements
    chop_amount = random.randint(0, len(original_list) // 10)
    modified_list = original_list[:-chop_amount] if chop_amount else original_list[:]
    # Randomly reverse the list
    if random.choice([True, False]):
        modified_list.reverse()
    # Add 1 to some random elements
    for i in range(random.randint(1, len(modified_list) // 10)):
        idx = random.randint(0, len(modified_list) - 1)
        if modified_list[idx] < max_height:
            modified_list[idx] += 1
    return modified_list

def run_program(program, input_data):
    result = subprocess.run(program, input=input_data, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
    if result.returncode != 0:
        print(f"Error running the program: {result.stderr}")
        return None
    return result.stdout.strip()

def generate_tests(config_file):
    with open(config_file, "r") as f:
        config = json.load(f)

    author_prog = "./author/author_fast"
    test_case_index = 1
    subtask_test_counts = []
    weights = []

    for subtask in config["subtasks"]:
        print(f"Generating tests for {subtask['name']}...")
        n_max = subtask["constraints"]["n_max"]
        max_height = subtask["constraints"]["max_height"]

        subtask_test_count = 0
        weights.append(subtask.get("weight", 20))  # Default weight if not provided

        for test_case in subtask["test_cases"]:
            num_cases = test_case["num_cases"]
            pattern = test_case.get("pattern")
            file_in = test_case.get("file_in")
            file_out = test_case.get("file_out")

            original_input = None
            original_output = None

            if file_in and file_out:
                # Load original input and output
                original_input = load_file_content(file_in)
                original_output = load_file_content(file_out)

            for i in range(num_cases):
                if i == num_cases - 1 and original_input and original_output:
                    # Use the original predefined test case
                    input_data = original_input
                    expected_output = original_output
                elif original_input and original_output:
                    # Generate a modified test case
                    modified_input_list = modify_test_case(original_input.split("\n")[1], max_height)
                    input_data = f"{len(modified_input_list)}\n" + " ".join(map(str, modified_input_list)) + "\n"
                    expected_output = run_program(author_prog, input_data)
                else:
                    # Generate test case dynamically
                    n = random.randint(1, n_max)
                    heights = generate_test_case(n, max_height, pattern)
                    input_data = f"{n}\n" + " ".join(map(str, heights)) + "\n"
                    expected_output = run_program(author_prog, input_data)

                if expected_output is None:
                    continue

                # Write input and output files
                input_path = f"tests/test_{test_case_index}.in"
                output_path = f"tests/test_{test_case_index}.out"
                with open(input_path, "w") as fin:
                    fin.write(input_data)
                with open(output_path, "w") as fout:
                    fout.write(expected_output)

                print(f"Generated test case {test_case_index} - Input and Output files created.")
                test_case_index += 1
                subtask_test_count += 1

        subtask_test_counts.append(subtask_test_count)

def generate_grade_properties(test_counts, weights, dependencies=None):
    groups = []
    start = 1
    for count in test_counts:
        end = start + count - 1
        groups.append(f"{start}-{end}")
        start = end + 1

    dependencies_str = dependencies if dependencies else "," * (len(groups) - 1)

    grade_properties = (
        f"time=1\n"
        f"memory=512\n"
        f"groups={','.join(groups)}\n"
        f"weights={','.join(map(str, weights))}\n"
        f"dependencies={dependencies_str}\n"
    )

    with open("grade.properties", "w") as f:
        f.write(grade_properties)

    print("Generated grade.properties with dependencies.")

def main():
    config_file = "config.json"
    if not os.path.exists(config_file):
        print(f"Error: {config_file} not found.")
        return

    with open(config_file, "r") as f:
        config = json.load(f)

    subtask_test_counts = []
    weights = []
    dependencies = []

    for subtask in config["subtasks"]:
        print(f"Processing {subtask['name']}...")
        weights.append(subtask.get("weight", 20))  # Default weight
        num_cases = sum(tc["num_cases"] for tc in subtask["test_cases"])
        subtask_test_counts.append(num_cases)
        dependencies.append(subtask.get("dependency", ""))  # Get dependency if specified

    # Build dependencies string
    dependencies_str = ",".join(";".join(map(str, dep)) if dep else "" for dep in dependencies)

    generate_tests(config_file)  # Generate test cases
    generate_grade_properties(subtask_test_counts, weights, dependencies_str)

if __name__ == "__main__":
    main()
